﻿--[[
	Copyright (c) 2012 Eyal Shilony

	Permission is hereby granted, free of charge, to any person obtaining
	a copy of this software and associated documentation files (the
	"Software"), to deal in the Software without restriction, including
	without limitation the rights to use, copy, modify, merge, publish,
	distribute, sublicense, and/or sell copies of the Software, and to
	permit persons to whom the Software is furnished to do so, subject to
	the following conditions:

	The above copyright notice and this permission notice shall be
	included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
	OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
	WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]

local addonName, addon = ...
local L = LibStub("AceLocale-3.0"):GetLocale(addonName)

local format = format
local select = select

local GetNumAddOns = GetNumAddOns
local GetAddOnMemoryUsage = GetAddOnMemoryUsage
local UpdateAddOnMemoryUsage = UpdateAddOnMemoryUsage
local GetNetStats = GetNetStats
local GetFramerate = GetFramerate

local function getTotalMemUsage()
	UpdateAddOnMemoryUsage()
	
	local long, short
	local totalMem = 0
	
	for i=1, GetNumAddOns(), 1 do
		totalMem = totalMem + GetAddOnMemoryUsage(i)
	end
	
	if totalMem > 0 then
		if totalMem > 1000 then
			totalMem = totalMem / 1000
			short = format("%.2f MB", totalMem)
			long = format(TOTAL_MEM_MB_ABBR, totalMem)
		else
			short = format("%.0f KB", totalMem)
			long = format(TOTAL_MEM_KB_ABBR, totalMem)
		end
	end
	
	return totalMem, long, short
end

local TopAddOns = { }
local numAddons = 0

function addon:DisplayTooltip(tooltip)
	local db = self.db.profile
	
	if numAddons ~= db.numAddons then
		for i=1, db.numAddons do
			TopAddOns[i] = { value = 0, name = "" }
		end
		numAddons = db.numAddons
	end

	local text = ""
	local i=0; j=0; k=0
	
	tooltip:ClearLines()
	
	tooltip:AddLine(self.name, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1)

	-- latency
	local _, _, latencyHome, latencyWorld = GetNetStats()
	text = format(MAINMENUBAR_LATENCY_LABEL, latencyHome, latencyWorld)
	tooltip:AddLine(" ")
	tooltip:AddLine(text, 1.0, 1.0, 1.0)
	if SHOW_NEWBIE_TIPS == "1" then
		tooltip:AddLine(NEWBIE_TOOLTIP_LATENCY, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1)
	end
	tooltip:AddLine(" ")
	
	-- framerate
	text = format(MAINMENUBAR_FPS_LABEL, GetFramerate());
	tooltip:AddLine(text, 1.0, 1.0, 1.0)
	if SHOW_NEWBIE_TIPS == "1" then
		tooltip:AddLine(NEWBIE_TOOLTIP_FRAMERATE, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1)
	end
	tooltip:AddLine(" ");

	text = format(MAINMENUBAR_BANDWIDTH_LABEL, GetAvailableBandwidth());
	tooltip:AddLine(text, 1.0, 1.0, 1.0);
	if SHOW_NEWBIE_TIPS == "1" then
		tooltip:AddLine(NEWBIE_TOOLTIP_BANDWIDTH, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1)
	end
	tooltip:AddLine(" ");

	local percent = floor(GetDownloadedPercentage() * 100 + 0.5);
	text = format(MAINMENUBAR_DOWNLOAD_PERCENT_LABEL, percent);
	tooltip:AddLine(text, 1.0, 1.0, 1.0);
	if ( SHOW_NEWBIE_TIPS == "1" ) then
		tooltip:AddLine(NEWBIE_TOOLTIP_DOWNLOAD_PERCENT, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1);
	end
	
	-- memory
	local totalMem, totalText = getTotalMemUsage()
	
	if totalMem > 0 then
		local mem
		
		for i = 1, db.numAddons, 1 do
			TopAddOns[i].value = 0
		end
		
		for i = 1, GetNumAddOns(), 1 do
			mem = GetAddOnMemoryUsage(i)
			for j = 1, db.numAddons, 1 do
				if mem > TopAddOns[j].value then
					for k = db.numAddons, 1, -1 do
						if k == j then
							TopAddOns[k].value = mem
							TopAddOns[k].name = GetAddOnInfo(i)
							break
						elseif k ~= 1 then
							TopAddOns[k].value = TopAddOns[k-1].value
							TopAddOns[k].name = TopAddOns[k-1].name
						end
					end
					break
				end
			end
		end
		
		tooltip:AddLine(" ")
		tooltip:AddLine(totalText, 1.0, 1.0, 1.0)
		if SHOW_NEWBIE_TIPS == "1" then
			tooltip:AddLine(NEWBIE_TOOLTIP_MEMORY, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1)
		end
		
		for i = 1, db.numAddons, 1 do
			if TopAddOns[i].value == 0 then
				break
			end
			
			size = TopAddOns[i].value
			
			if size > 1000 then
				size = size / 1000
				text = format(ADDON_MEM_MB_ABBR, size, TopAddOns[i].name)
			else
				text = format(ADDON_MEM_KB_ABBR, size, TopAddOns[i].name)
			end
			
			tooltip:AddLine(text, 1.0, 1.0, 1.0)
		end
	end

	tooltip:Show()
end

local function getAvailableSlots()
	local totalSlots = 0
	local freeSlots = 0
	local usedSlots = 0
	
	for i = 0, NUM_BAG_SLOTS do
		totalSlots = totalSlots + GetContainerNumSlots(i)
		freeSlots = freeSlots + GetContainerNumFreeSlots(i)
	end
	
	usedSlots = totalSlots - freeSlots
	
	return totalSlots, freeSlots, usedSlots
end

local display = {
	["@latency"] = function() 
		return format("%d ms", select(3, GetNetStats())) 
	end,
	["@fps"] = function() 
		return format("%.0f fps", GetFramerate()) 
	end,
	["@memory"] = function() 
		return select(3, getTotalMemUsage()) 
	end,
	["@bags"] = function() 
		local totalSlots, totalFreeSlots = getAvailableSlots()
		return ("%s/%s"):format(totalFreeSlots, totalSlots)
	end
}

function addon:GetTitle()
	if not self.db.profile.display or not (type(self.db.profile.display) == "string") then
		self.db.profile.display = self.db.defaults.profile.display
	end
	
	local title = self.db.profile.display and self.db.profile.display:gsub("@%w+", function(key)
		local value = display[key];
		return value and value()
	end)
	
	return title
end